"""tree.py @ pyckage.pyckagelib
"""
import os
import abc
import pathlib

from .data import PackageData
from .tracker import FileTracker
from .template import FileTemplate


class FileTree(object, metaclass=abc.ABCMeta):

    DIR = "DIR"
    FILE = "FILE"

    FILE_TYPES = {FILE, DIR, None}

    __datum__ = {}

    @classmethod
    def set(self, **kwargs):
        self.__datum__.update(kwargs)

    @classmethod
    def get(self, key: str):
        return self.__datum__[key]

    def __init__(
        self,
        type_: str,
        head: str = None,
        tail: list = None,
        *,
        cond: bool = True,
        template: FileTemplate = None,
    ):
        """"""
        self.type = type_
        self.head = head
        self.cond = cond

        if tail is None:
            self.tail = []
        else:
            self.tail = [subtree for subtree in tail if subtree.cond]

        self.template = template

    def __repr__(self, level: int = 0) -> str:
        """"""
        if self.tail:
            return (
                f"FileTree({self.type}, {self.head!r}, ["
                + "\n"
                + ",\n".join(
                    (
                        "\t" * (level + 1)
                        + f"{item.__repr__(level+1) if type(item) is type(self) else repr(item)}"
                        for item in self.tail
                    )
                )
                + "\n"
                + "\t" * level
            )
        else:
            return f"FileTree({self.type}, {self.head!r})"

    @classmethod
    def root(cls, tail: list):
        """"""
        return cls(None, None, tail)

    def grow(self, root_path: pathlib.Path, *, tracker: FileTracker = None):
        """"""
        if tracker is None:
            raise ValueError("`tracker` is None.")

        if self.type is not None:
            if not self.cond:
                return

            path = root_path.joinpath(self.head)

            tracker.create(self.type, path, template=self.template)
        else:
            path = root_path

        if self.tail is not None:
            for subtree in self.tail:
                subtree.grow(path, tracker=tracker)


import os
import json
import argparse

from ..pyckage import Pyckage, _PYCKAGE_DATA as package_data
from ..pyckagelib.error import PyckageError


def pack(args: argparse.Namespace):
    """"""
    pyckage = Pyckage.pack(args)
    pyckage.save()


import os
import re
import sys
import site
import json
import argparse
import platform
import configparser
from pathlib import Path
from contextlib import contextmanager

## Exceptions
from .pyckagelib.error import PyckageError, PyckageExit
from .pyckagelib import (
    PackageData,
    FileTree,
    FileTemplate,
    FileTracker,
    Validate,
)

__version__ = "0.1.1"

## To be used with the module
_PYCKAGE_DATA = PackageData("pyckage")


class PyckageValidate(Validate):
    """"""

    RE_FLAGS = re.UNICODE

    RE_AUTHOR = re.compile(r"[\S ]+", RE_FLAGS)
    RE_USER = re.compile(r"[a-zA-Z]+", RE_FLAGS)
    RE_EMAIL = re.compile(
        r"^(([^<>()[\]\.,;:\s@\"]+(\.[^<>()[\]\.,;:\s@\"]+)*)|(\".+\"))@(([^<>()[\]\.,;:\s@\"]+\.)+[^<>()[\]\.,;:\s@\"]{2,})$",
        RE_FLAGS,
    )
    RE_VERSION = re.compile(
        r"^(0|[1-9]\d*)(\.(0|[1-9]\d*))(\.(0|[1-9]\d*))?$", RE_FLAGS
    )
    RE_PACKAGE = re.compile(r"^[a-zA-Z][a-zA-Z\_\-]+$", RE_FLAGS)
    RE_DESCRIPTION = re.compile(r"[\S\s]*", RE_FLAGS | re.MULTILINE)

    @classmethod
    def description(cls, description: str, null: bool = None) -> str:
        return cls._regex(cls.RE_DESCRIPTION, description, "description", null=null)

    @classmethod
    def author(cls, author: str, null: object = None) -> str:
        return cls._regex(cls.RE_AUTHOR, author, "author", null=null)

    @classmethod
    def user(cls, user: str, null: object = None) -> str:
        return cls._regex(cls.RE_AUTHOR, user, "user", null=null)

    @classmethod
    def email(cls, email: str, null: bool = None) -> str:
        return cls._regex(cls.RE_EMAIL, email, "email", null=null)

    @classmethod
    def version(cls, version: str, null: bool = False) -> str:
        return cls._regex(cls.RE_VERSION, version, "version", null=null)

    @classmethod
    def package(cls, package: str, null: bool = False) -> str:
        return cls._regex(cls.RE_PACKAGE, package, "package name", null=null)

    @classmethod
    def path(cls, path: str, null: bool = None) -> Path:
        if path is None and null is not False:
            return Path.cwd().absolute()
        elif type(path) is str or type(path) is Path:
            path = Path(path)
            if not path.exists():
                raise cls.Invalid(f"Path `{path}` not found.")
            else:
                return path.absolute()
        else:
            raise cls.Invalid(f"Invalid path type {type(path)}")

    @classmethod
    def platform(cls, platform_: str, null: bool = False) -> str:
        if platform_ is None and null is not False:
            return platform.system()
        elif type(platform_) is str:
            if platform_ in {"Windows", "Linux", "Darwin", "Java", "Solaris"}:
                return platform_
            else:
                raise cls.Invalid(f"Invalid platform {platform_}")
        else:
            raise cls.Invalid(f"Invalid platform type {type(platform_)}")


class PyckageTemplate(FileTemplate):
    """"""

    package_data = _PYCKAGE_DATA

    @FileTemplate.template("setup.cfg")
    def setupcfg(self, source: str, dest: str) -> str:
        cfg_files = ["setup-plain.cfg.t"]

        if self.get("args").data:
            cfg_files.append("setup-data.cfg.t")

        if self.get("args").script:
            cfg_files.append("setup-script.cfg.t")

        if self.get("args").github:
            cfg_files.append("setup-github.cfg.t")

        ## Parse all configuration templates.
        parser = configparser.ConfigParser()
        parser.read([self.package_data.get_data_path(fname) for fname in cfg_files])

        ## Summarise and write to file.
        with self.package_data.open_data("setup.cfg.t", mode="w") as file:
            parser.write(file)

        return (self.package_data.get_data_path("setup.cfg.t"), Path("setup.cfg"))


class Pyckage(object):
    """"""

    package_data = _PYCKAGE_DATA

    def __init__(
        self,
        *,
        description: str = None,
        package: str = None,
        pathlog: list = None,
        version: str = None,
        author: str = None,
        email: str = None,
        user: str = None,
        path: str = None,
        meta: dict = None,
    ):
        self.description = description
        self.package = package
        self.version = version
        self.author = author
        self.email = email
        self.user = user
        self.path = path
        self.meta = meta

        # Meta parameters
        self._pyckage_version = self.kwget("pyckage_version", self.meta, __version__)
        self._system_platform = self.kwget(
            "system_platform", self.meta, platform.system()
        )

        self._py_min = self.kwget(
            "py_min", self.meta, f"{sys.version_info.major}.{sys.version_info.minor}"
        )
        self._py_max = self.kwget("py_max", self.meta, f"{sys.version_info.major + 1}")

        self.validate()

        self.tracker = FileTracker(pathlog)

    @property
    def info(self) -> dict:
        return {
            ## External
            "package_data": self.package_data.package_data,
            ## Common
            "description": self.description,
            "package": self.package,
            "version": self.version,
            "author": self.author,
            "email": self.email,
            "user": self.user,
            ## Meta
            "py_min": self._py_min,
            "py_max": self._py_max,
        }

    @classmethod
    def kwget(cls, key: object, kwargs: dict, default: object = None):
        if kwargs is None or key not in kwargs:
            return default
        else:
            return kwargs[key]

    def validate(self):
        try:
            # description
            self.description = PyckageValidate.description(self.description)

            # package
            self.package = PyckageValidate.package(self.package)

            # version
            self.version = PyckageValidate.version(self.version)

            # author
            self.author = PyckageValidate.author(self.author)

            # email
            self.email = PyckageValidate.email(self.email)

            # user
            self.user = PyckageValidate.user(self.user)

            self.validate_meta()
        except PyckageValidate.Invalid as error:
            raise ValueError(*error.args)

    def validate_meta(self):
        """"""
        self._pyckage_version = PyckageValidate.version(self._pyckage_version)
        self._system_platform = PyckageValidate.platform(self._system_platform)

    @classmethod
    def load(cls, pyckage_path: str):
        path = Path(pyckage_path).joinpath(".pyckage").absolute()

        if not path.exists():
            raise FileNotFoundError(f"No Pyckage installed at `{pyckage_path}`.")

        with open(path, mode="r") as file:
            return cls.JSON_DECODE(file.read())

    def save(self):
        """"""
        if not self.path.exists():
            raise FileNotFoundError(
                f"Misleading pyckage installation for `{self.package}`. Try running `pyckage fix`."
            )

        path = self.path.joinpath(".pyckage")

        with open(path, mode="w") as file:
            return file.write(self.JSON_ENCODE())

    @classmethod
    def pack(cls, args: argparse.Namespace):
        """"""
        ## Get pyckage path
        path = PyckageValidate.path(args.path)

        if not path.exists():
            raise FileNotFoundError(f"No target directory `{path}`.")
        elif not path.is_dir():
            raise OSError(f"`{path}` is not a directory.")

        pyckage_path = path.joinpath(".pyckage")

        if pyckage_path.exists():
            raise FileExistsError(
                "There is an existing Pyckage in this folder. Try running `pyckage update` or `pyckage fix`"
            )

        package: str = args.package

        if package is None:
            package = path.stem

        version: str = args.version

        if version is None:
            version = "0.0.0"

        ## Get global config
        config = cls.package_data.get_config()

        author: str = args.author

        if author is None:
            author = config["author"]

        email: str = args.email

        if email is None:
            email = config["email"]

        user: str = args.user

        if user is None:
            user = config["user"]

        description: str = args.description

        pyckage = cls(
            package=package,
            version=version,
            author=author,
            email=email,
            user=user,
            path=path,
            description=description,
        )

        pyckage.tree(args=args).grow(pyckage.path, tracker=pyckage.tracker)

        return pyckage

    def JSON_ENCODE(self) -> str:
        """
        {
            'meta': {
                    'pyckage_version': '0.0.0',
                    'platform': 'windows',
                    'pathlog': ['c:/users/author/pyckage/*', ...],
                    'py_min': '3.7',
                    'py_max': '4',
                },
            'package': 'pyckage',
            'version': '1.0.0',
            'author': 'Author Smith',
            'email': 'author@email.net',
            'user': 'author1998',
            'path': 'c:/users/author/pyckage/',
            'description': 'This is clearly a Python Package'
        }
        """
        return json.dumps(
            {
                "meta": {
                    "pyckage_version": self._pyckage_version,
                    "system_platform": self._system_platform,
                    "py_min": self._py_min,
                    "py_max": self._py_max,
                },
                "description": self.description,
                "pathlog": [[type_, str(path)] for type_, path in self.tracker.pathlog],
                "package": self.package,
                "version": self.version,
                "author": self.author,
                "email": self.email,
                "user": self.user,
                "path": str(self.path),
            }
        )

    @classmethod
    def JSON_DECODE(cls, json_data: str):
        return cls(**json.loads(json_data))

    def clear(self):
        """"""
        self.tracker.clear()

    # Some Checks
    @classmethod
    def has_pyckage(cls, path: str) -> bool:
        return Path(path).joinpath(".pyckage").exists()

    @classmethod
    def exit(cls, code: int = 0, msg: str = ""):
        raise PyckageExit(code, msg)

    def tree(self, *, args: argparse.Namespace = None) -> FileTree:
        """"""
        if args is None:
            raise ValueError("`args` is None")
        else:
            PyckageTemplate.set(args=args)

        T = FileTree  # Alias

        return T.root(
            [
                T(
                    T.DIR,
                    "src",
                    [
                        T(
                            T.DIR,
                            self.package,
                            [
                                T(
                                    T.FILE,
                                    "__init__.py",
                                    template=PyckageTemplate(
                                        source=self.package_data.get_data_path(
                                            "__init__.py.t"
                                        ),
                                        info=self.info,
                                    ),
                                ),
                            ],
                        ),
                    ],
                ),
                T(
                    T.DIR,
                    "bin",
                    [
                        T(
                            T.FILE,
                            self.package,
                            template=PyckageTemplate(
                                source=self.package_data.get_data_path("package.t"),
                                info=self.info,
                            ),
                        )
                    ],
                    cond=args.script,
                ),
                T(T.DIR, "docs", []),
                T(T.DIR, "data", [], cond=args.data),
                T(
                    T.FILE,
                    "setup.py",
                    template=PyckageTemplate(
                        source=self.package_data.get_data_path("setup.py.t"),
                        info=self.info,
                    ),
                ),
                T(
                    T.FILE,
                    "setup.cfg",
                    template=PyckageTemplate(
                        source=self.package_data.get_data_path("setup.cfg.t"),
                        info=self.info,
                        key="setup.cfg",
                    ),
                ),
                T(T.FILE, ".pyckage"),
            ]
        )


import argparse
from ..pyckage import Pyckage, PyckageValidate


def clear(args: argparse.Namespace):
    """"""
    path = PyckageValidate.path(args.path)

    pyckage = Pyckage.load(path)
    pyckage.clear()


"""template.py @ pyckage.pyckagelib

This utility allows for file template generation, applying changes from a given dictionary.
"""
from pathlib import Path


class FileFormatter(dict):
    """"""

    def __getitem__(self, key):
        value = dict.__getitem__(self, key)
        return f"{{{key}}}" if value is None else value

    def __missing__(self, key):
        return f"{{{key}}}"


class FileTemplate:
    """"""

    FILE = "FILE"

    __hooks__ = {}
    __datum__ = {}

    def __init__(self, source: Path, *, key: str = None, info: dict = None):
        self.source = source
        self.info = info
        self.key = key

    @classmethod
    def set(self, **kwargs):
        self.__datum__.update(kwargs)

    @classmethod
    def get(self, key: str):
        return self.__datum__[key]

    @classmethod
    def template(cls, key: str):
        """Registers new template hook. A template hook is a bound method.

        Every hook is recorded under a key. Calling a template's make function with the `template` keyword argument will use `template` as key for loading the hook. Just as said, using the `args` and `kwargs` keyword arguments will tell what to pass for the hook.

        hook's return is a string telling from which file to read the template, i.e., overrides `from_` normal behavior. Thus, a regular use case would be to create a new template file from the usual one and use it right away. This allows for advanced template building. You might want to make use of the `tempfile` module.

        Parameters
        ----------
        key: str
            hook registry key
        """

        def decor(callback):
            cls.__hooks__[key] = callback
            return callback

        return decor

    def __default_hook(self, from_: Path, to_: Path) -> (str, str):
        """Default behavior is to look at the package data folder."""
        return (from_, to_)

    def make(self, path: Path):
        """"""
        ## Retrieve path information
        if self.key in self.__hooks__:
            source_path, dest_path = self.__hooks__[self.key](self, self.source, path)
        else:
            source_path, dest_path = self.__default_hook(self.source, path)

        ## Read template file
        with open(source_path, mode="r") as r_file:
            content = r_file.read()

        with open(dest_path, "w") as w_file:
            if self.info is None:
                w_file.write(content)
            else:
                w_file.write(content.format_map(FileFormatter(self.info)))


__all__ = ["FileTemplate"]


"""`data.py` @ pyckage.pyckagelib

This module provides the basic interfaces for reading/writing files hosted at a pyckage dat folder.

Other modules are heavily dependant in this one, thus it must not have any dependencies besides the standard library.

STATUS: COMPLETE
"""
import _io
import os
import sys
import site
import json

from pathlib import Path


class PackageData:
    """"""

    USR = site.USER_BASE
    SYS = sys.prefix

    __ref__ = {}

    def __new__(cls, package: str):
        """"""
        if package not in cls.__ref__:
            cls.__ref__[package] = object.__new__(cls)
        return cls.__ref__[package]

    def __init__(self, package: str):
        """"""
        self.package = package

    @property
    def package_data(self):
        return f"{self.package}_data"

    def get_config(self) -> dict:
        """"""
        with self.open_data(f".{self.package}-config", mode="r") as file:
            return json.load(file)

    def set_config(self, config: dict):
        """"""
        with self.open_data(f".{self.package}-config", mode="w") as file:
            json.dump(config, file)

    def get_data_path(self, fname: str) -> Path:
        """

        Examples
        --------
        >>> PyckageData.get_data_path('') # retrieves base path
        """

        if self.package_data is not None:
            sys_path = Path(self.SYS).joinpath(self.package_data, fname)
            if not os.path.exists(sys_path):
                usr_path = Path(self.USR).joinpath(self.package_data, fname)
                if not os.path.exists(usr_path):
                    if fname:
                        raise FileNotFoundError(
                            f"File {fname} not installed in {self.package_data}."
                        )
                    else:
                        raise FileNotFoundError(
                            f"Package data `{self.package_data}` not installed."
                        )
                else:
                    return usr_path
            else:
                return sys_path
        else:
            raise NameError("PycakgeData.PACKAGE_DATA is not defined.")

    def open_data(
        self, fname: str, mode: str = "r", *args: tuple, **kwargs: dict
    ) -> _io.TextIOWrapper:
        """Opens file relative to package data folder.

        Parameters
        ----------
        fname: str
            File to open.
        mode: str
            File read mode, e.g., 'r', 'w', 'rb', 'wb', etc.
        *args: tuple
            Extra arguments for passing to open(...)
        **kwargs: dict
            Extra keyword-arguments for passing to open(...)
        """
        if self.package_data is None:
            raise NameError("PycakgeData.package_data is not defined.")
        else:
            path = self.get_data_path("")
            return open(
                path.joinpath(fname),
                mode=mode,
                *args,
                **kwargs,
            )

    def open_local(
        self, fname: str, *, path: str = None, mode: str = "r"
    ) -> _io.TextIOWrapper:
        """"""
        if path is None:
            return self.open_data(fname, mode=mode)
        else:
            path = Path(path)
            return open(path.joinpath(fname), mode=mode)


__all__ = ["PackageData"]


"""tracker.py @ pyckage.pyckagelib

"""
import os
import sys
from pathlib import Path

from .template import FileTemplate


class FileTracker:
    """"""

    DIR = "DIR"
    FILE = "FILE"

    FILE_TYPES = {DIR, FILE}

    def __init__(self, pathlog: list = None):
        self.pathlog = (
            [(type_, Path(path)) for type_, path in pathlog]
            if pathlog is not None
            else []
        )

    @staticmethod
    def touch(path: str):
        """"""
        with open(path, "a"):
            pass

    def clear(self):
        """"""
        while self.pathlog:
            type_, path = self.pathlog.pop()
            try:
                if type_ == self.FILE:
                    os.remove(path)
                    print(f"Remove FILE `{path}`.")
                elif type_ == self.DIR:
                    os.rmdir(path)
                    print(f"Remove DIR `{path}`.")
                else:
                    raise TypeError
            except:
                raise EnvironmentError(
                    "Fatal Error: run `pyckage fix` or try fixing package contents manually."
                )

    def create(self, type_: str, path: Path, template: FileTemplate = None):
        """"""
        if path.exists():
            raise FileExistsError(f"`{path}` already exists.")
        elif type_ in self.FILE_TYPES:
            if type_ == self.DIR:
                print(f"Create FILE `{path}`.")
                os.mkdir(path)
                self.pathlog.append((type_, path))

            elif type_ == self.FILE:
                print(f"Create DIR `{path}`.")
                if template is None:
                    self.touch(path)
                else:
                    template.make(path=path)
                self.pathlog.append((type_, path))

            else:
                raise TypeError(f"Invalid type {type_}.")
        else:
            raise TypeError(f"Invalid type {type_}.")


__all__ = ["FileTracker"]


"""`error.py` @ pyckage.pyckagelib

STATUS: COMPLETE
"""


class PyckageError(Exception):
    """"""


class PyckageExit(Exception):
    """"""

    def __init__(self, code: int = 0, *args: tuple, **kwargs: dict):
        Exception.__init__(self, *args, **kwargs)
        self.code = code


""" Pyckage - A Python package utility.
"""
import argparse
from ..pyckage import Pyckage
from ..pyckagelib.error import PyckageExit, PyckageError

## Commands
from .pack import pack
from .clear import clear
from .config import config


class PyckageParser(argparse.ArgumentParser):
    pass


def main():
    kwargs = {
        "prog": "pyckage",
        "description": __doc__,
    }

    parser = PyckageParser(**kwargs)
    parser.set_defaults(func=None)

    subparsers = parser.add_subparsers()

    ## pack
    pack_parser = subparsers.add_parser("pack")
    pack_parser.add_argument("path", type=str, nargs="?", help="packaging path.")

    # Package info
    pack_parser.add_argument(
        "--package", dest="package", type=str, action="store", help="project name."
    )
    pack_parser.add_argument(
        "--version", dest="version", type=str, action="store", help="project version."
    )
    pack_parser.add_argument(
        "--description",
        dest="description",
        type=str,
        action="store",
        help="project description.",
    )
    pack_parser.add_argument(
        "--author",
        dest="author",
        type=str,
        action="store",
        help="project author.",
    )
    pack_parser.add_argument(
        "--email",
        dest="email",
        type=str,
        action="store",
        help="project email.",
    )
    pack_parser.add_argument(
        "--user",
        dest="user",
        type=str,
        action="store",
        help="project user.",
    )

    # Conditional options
    pack_parser.add_argument(
        "-d", "--data", action="store_true", help="whether to store external data."
    )
    pack_parser.add_argument(
        "-s", "--script", action="store_true", help="whether to deploy cli script."
    )
    pack_parser.add_argument(
        "-g", "--github", action="store_true", help="whether to add github info."
    )
    pack_parser.set_defaults(func=pack)

    ## clear
    clear_parser = subparsers.add_parser("clear")
    clear_parser.add_argument("path", type=str, nargs="?", help="packaging path.")
    clear_parser.set_defaults(func=clear)

    ## config
    config_parser = subparsers.add_parser("config")

    config_group = config_parser.add_mutually_exclusive_group()
    config_group.add_argument("path", type=str, nargs="?", help="packaging path.")
    config_group.add_argument(
        "--global",
        dest="global_",
        action="store_true",
        help="applies global config to pyckages.",
    )

    config_parser.add_argument("-a", "--author", help="sets default author name.")
    config_parser.add_argument("-e", "--email", help="sets default email address.")
    config_parser.add_argument("-u", "--user", help="sets default username (github).")
    config_parser.set_defaults(func=config)

    ## run
    args = parser.parse_args()
    try:
        if args.func is not None:
            args.func(args)
        else:
            parser.print_help()
    except PyckageExit as exit_:
        exit(exit_.code)
    else:
        exit(0)


if __name__ == "__main__":
    main()


import pathlib
import argparse

from ..pyckage import Pyckage, PyckageValidate, _PYCKAGE_DATA as package_data


def config(args: argparse.Namespace):
    if args.global_:
        config_global(args)
    else:
        config_local(args)


def config_local(args: argparse.Namespace):
    """"""
    if args.path is None:
        path = pathlib.Path.cwd().absolute()
    else:
        path = pathlib.Path(args.path).absolute()

    if not Pyckage.has_pyckage(path):
        print(f"No Pyckage installed at `{args.path}`.")
        return

    pyckage = Pyckage.load(path)

    author = PyckageValidate.author(args.author)

    if author is not None:
        pyckage.author = author

    email = PyckageValidate.email(args.email)

    if email is not None:
        pyckage.email = email

    user = PyckageValidate.user(args.user)

    if user is not None:
        pyckage.user = user

    config_fields = {author, email, user}

    if any(map(lambda x: x is not None, config_fields)):
        Pyckage.save(pyckage)
    else:
        print(f'author = {pyckage.author}')
        print(f'email = {pyckage.email}')
        print(f'user = {pyckage.user}')

def config_global(args: argparse.Namespace):
    """"""
    config = {
        "author": PyckageValidate.author(args.author),
        "email": PyckageValidate.email(args.email),
        "user": PyckageValidate.user(args.user),
    }

    updates = {k: v for k, v in config.items() if v is not None}

    config = package_data.get_config()

    if not updates:
        for k, v in config.items():
            print(f"{k} = {v}")
    else:
        for k, v in updates.items():
            config[k] = v

        package_data.set_config(config)


"""`pyckagelib.py` @ pyckage.pyckagelib
"""
import re


class Validate:
    """"""

    class Invalid(Exception):
        pass

    @classmethod
    def _regex(cls, regex: re.Pattern, s: str, field: str = "", null: bool = False):
        if s is None and null is not False:
            return null
        elif type(s) is not str or regex.match(s) is None:
            raise cls.Invalid(f"Invalid {field}: {s}")
        else:
            return s


__all__ = ["Validate"]
